/*
 * remux.c: Tools for detecting frames and handling PAT/PMT
 *
 * See the main source file 'vdr.c' for copyright information and
 * how to reach the author.
 *
 * $Id: remux.c 2.47 2010/06/05 13:32:15 kls Exp $
 */

#include "remux.h"
//#include "device.h"
#include "libsi/si.h"
#include "libsi/section.h"
#include "libsi/descriptor.h"
//#include "shutdown.h"
//#include "tools.h"

// Set these to 'true' for debug output:
static bool DebugPatPmt = false;
//static bool DebugFrames = false;

#define dbgpatpmt(a...) if (DebugPatPmt) fprintf(stderr, a)
#define dbgframes(a...) if (DebugFrames) fprintf(stderr, a)


#define VIDEO_STREAM_S   0xE0

// --- Some TS handling tools ------------------------------------------------

int64_t TsGetPts(const uchar *p, int l)
{
  // Find the first packet with a PTS and use it:
  while (l > 0) {
        const uchar *d = p;
        if (TsPayloadStart(d) && TsGetPayload(&d) && PesHasPts(d))
           return PesGetPts(d);
        p += TS_SIZE;
        l -= TS_SIZE;
        }
  return -1;
}

void TsSetTeiOnBrokenPackets(uchar *p, int l)
{
  bool Processed[MAXPID] = { false };
  while (l >= TS_SIZE) {
        if (*p != TS_SYNC_BYTE)
           break;
        int Pid = TsPid(p);
        if (!Processed[Pid]) {
           if (!TsPayloadStart(p))
              p[1] |= TS_ERROR;
           else
              Processed[Pid] = true;
           }
        l -= TS_SIZE;
        p += TS_SIZE;
        }
}

void TsExtendAdaptionField(unsigned char *Packet, int ToLength)
{
    // Hint: ExtenAdaptionField(p, TsPayloadOffset(p) - 4) is a null operation

    int Offset = TsPayloadOffset(Packet); // First byte after existing adaption field

    if (ToLength <= 0)
    {
        // Remove adaption field
        Packet[3] = Packet[3] & ~TS_ADAPT_FIELD_EXISTS;
        return;
    }

    // Set adaption field present
    Packet[3] = Packet[3] | TS_ADAPT_FIELD_EXISTS;

    // Set new length of adaption field:
    Packet[4] = ToLength <= TS_SIZE-4 ? ToLength-1 : TS_SIZE-4-1;

    if (Packet[4] == TS_SIZE-4-1)
    {
        // No more payload, remove payload flag
        Packet[3] = Packet[3] & ~TS_PAYLOAD_EXISTS;
    }

    int NewPayload = TsPayloadOffset(Packet); // First byte after new adaption field

    // Fill new adaption field
    if (Offset == 4 && Offset < NewPayload)
        Offset++; // skip adaptation_field_length
    if (Offset == 5 && Offset < NewPayload)
        Packet[Offset++] = 0; // various flags set to 0
    while (Offset < NewPayload)
        Packet[Offset++] = 0xff; // stuffing byte
}

// --- cPatPmtParser ---------------------------------------------------------

cPatPmtParser::cPatPmtParser(bool UpdatePrimaryDevice)
{
//  updatePrimaryDevice = UpdatePrimaryDevice;
  Reset();
}

void cPatPmtParser::Reset(void)
{
  pmtSize = 0;
  patVersion = pmtVersion = -1;
  pmtPid = -1;
  vpid = vtype = 0;
  ppid = 0;
}

void cPatPmtParser::ParsePat(const uchar *Data, int Length)
{
  // Unpack the TS packet:
  int PayloadOffset = TsPayloadOffset(Data);
  Data += PayloadOffset;
  Length -= PayloadOffset;
  // The PAT is always assumed to fit into a single TS packet
  if ((Length -= Data[0] + 1) <= 0)
     return;
  Data += Data[0] + 1; // process pointer_field
  SI::PAT Pat(Data, false);
  if (Pat.CheckCRCAndParse()) {
     dbgpatpmt("PAT: TSid = %d, c/n = %d, v = %d, s = %d, ls = %d\n", Pat.getTransportStreamId(), Pat.getCurrentNextIndicator(), Pat.getVersionNumber(), Pat.getSectionNumber(), Pat.getLastSectionNumber());
     if (patVersion == Pat.getVersionNumber())
        return;
     SI::PAT::Association assoc;
     for (SI::Loop::Iterator it; Pat.associationLoop.getNext(assoc, it); ) {
         dbgpatpmt("     isNITPid = %d\n", assoc.isNITPid());
         if (!assoc.isNITPid()) {
            pmtPid = assoc.getPid();
            dbgpatpmt("     service id = %d, pid = %d\n", assoc.getServiceId(), assoc.getPid());
            }
         }
     patVersion = Pat.getVersionNumber();
     }
  else
     esyslog("ERROR: can't parse PAT");
}

void cPatPmtParser::ParsePmt(const uchar *Data, int Length)
{
  // Unpack the TS packet:
  bool PayloadStart = TsPayloadStart(Data);
  int PayloadOffset = TsPayloadOffset(Data);
  Data += PayloadOffset;
  Length -= PayloadOffset;
  // The PMT may extend over several TS packets, so we need to assemble them
  if (PayloadStart) {
     pmtSize = 0;
     if ((Length -= Data[0] + 1) <= 0)
        return;
     Data += Data[0] + 1; // this is the first packet
     if (SectionLength(Data, Length) > Length) {
        if (Length <= int(sizeof(pmt))) {
           memcpy(pmt, Data, Length);
           pmtSize = Length;
           }
        else
           esyslog("ERROR: PMT packet length too big (%d byte)!", Length);
        return;
        }
     // the packet contains the entire PMT section, so we run into the actual parsing
     }
  else if (pmtSize > 0) {
     // this is a following packet, so we add it to the pmt storage
     if (Length <= int(sizeof(pmt)) - pmtSize) {
        memcpy(pmt + pmtSize, Data, Length);
        pmtSize += Length;
        }
     else {
        esyslog("ERROR: PMT section length too big (%d byte)!", pmtSize + Length);
        pmtSize = 0;
        }
     if (SectionLength(pmt, pmtSize) > pmtSize)
        return; // more packets to come
     // the PMT section is now complete, so we run into the actual parsing
     Data = pmt;
     }
  else
     return; // fragment of broken packet - ignore
  SI::PMT Pmt(Data, false);
  if (Pmt.CheckCRCAndParse()) {
     dbgpatpmt("PMT: sid = %d, c/n = %d, v = %d, s = %d, ls = %d\n", Pmt.getServiceId(), Pmt.getCurrentNextIndicator(), Pmt.getVersionNumber(), Pmt.getSectionNumber(), Pmt.getLastSectionNumber());
     dbgpatpmt("     pcr = %d\n", Pmt.getPCRPid());
     if (pmtVersion == Pmt.getVersionNumber())
        return;
//     if (updatePrimaryDevice)
//        cDevice::PrimaryDevice()->ClrAvailableTracks(false, true);
     int NumApids = 0;
     int NumDpids = 0;
     int NumSpids = 0;
     vpid = vtype = 0;
     ppid = 0;
     apids[0] = 0;
     dpids[0] = 0;
     spids[0] = 0;
     atypes[0] = 0;
     dtypes[0] = 0;
     SI::PMT::Stream stream;
     for (SI::Loop::Iterator it; Pmt.streamLoop.getNext(stream, it); ) {
         dbgpatpmt("     stream type = %02X, pid = %d", stream.getStreamType(), stream.getPid());
         switch (stream.getStreamType()) {
           case 0x01: // STREAMTYPE_11172_VIDEO
           case 0x02: // STREAMTYPE_13818_VIDEO
           case 0x1B: // MPEG4
                      vpid = stream.getPid();
                      vtype = stream.getStreamType();
                      ppid = Pmt.getPCRPid();
                      break;
           case 0x03: // STREAMTYPE_11172_AUDIO
           case 0x04: // STREAMTYPE_13818_AUDIO
           case 0x0F: // ISO/IEC 13818-7 Audio with ADTS transport sytax
           case 0x11: // ISO/IEC 14496-3 Audio with LATM transport syntax
                      {
                      if (NumApids < MAXAPIDS) {
                         apids[NumApids] = stream.getPid();
                         atypes[NumApids] = stream.getStreamType();
/*
                         *alangs[NumApids] = 0;
                         SI::Descriptor *d;
                         for (SI::Loop::Iterator it; (d = stream.streamDescriptors.getNext(it)); ) {
                             switch (d->getDescriptorTag()) {
                               case SI::ISO639LanguageDescriptorTag: {
                                    SI::ISO639LanguageDescriptor *ld = (SI::ISO639LanguageDescriptor *)d;
                                    SI::ISO639LanguageDescriptor::Language l;
                                    char *s = alangs[NumApids];
                                    int n = 0;
                                    for (SI::Loop::Iterator it; ld->languageLoop.getNext(l, it); ) {
                                        if (*ld->languageCode != '-') { // some use "---" to indicate "none"
                                           dbgpatpmt(" '%s'", l.languageCode);
                                           if (n > 0)
                                              *s++ = '+';
                                           strn0cpy(s, I18nNormalizeLanguageCode(l.languageCode), MAXLANGCODE1);
                                           s += strlen(s);
                                           if (n++ > 1)
                                              break;
                                           }
                                        }
                                    }
                                    break;
                               default: ;
                               }
                             delete d;
                             }
                         if (updatePrimaryDevice)
                            cDevice::PrimaryDevice()->SetAvailableTrack(ttAudio, NumApids, apids[NumApids], alangs[NumApids]);
*/
                         NumApids++;
                         apids[NumApids]= 0;
                         }
                      }
                      break;
           case 0x06: // STREAMTYPE_13818_PES_PRIVATE
                      {
                      int dpid = 0;
                      int dtype = 0;
//                      char lang[MAXLANGCODE1] = "";
                      SI::Descriptor *d;
                      for (SI::Loop::Iterator it; (d = stream.streamDescriptors.getNext(it)); ) {
                          switch (d->getDescriptorTag()) {
                            case SI::AC3DescriptorTag:
                            case SI::EnhancedAC3DescriptorTag:
                                 dbgpatpmt(" AC3");
                                 dpid = stream.getPid();
                                 dtype = d->getDescriptorTag();
                                 break;
                            case SI::SubtitlingDescriptorTag:
                                 dbgpatpmt(" subtitling");
                                 if (NumSpids < MAXSPIDS) {
                                    spids[NumSpids] = stream.getPid();
//                                    *slangs[NumSpids] = 0;
                                    subtitlingTypes[NumSpids] = 0;
                                    compositionPageIds[NumSpids] = 0;
                                    ancillaryPageIds[NumSpids] = 0;
//                                    SI::SubtitlingDescriptor *sd = (SI::SubtitlingDescriptor *)d;
                                    SI::SubtitlingDescriptor::Subtitling sub;
/*
                                    char *s = slangs[NumSpids];
                                    int n = 0;
                                    for (SI::Loop::Iterator it; sd->subtitlingLoop.getNext(sub, it); ) {
                                        if (sub.languageCode[0]) {
                                           dbgpatpmt(" '%s'", sub.languageCode);
                                           subtitlingTypes[NumSpids] = sub.getSubtitlingType();
                                           compositionPageIds[NumSpids] = sub.getCompositionPageId();
                                           ancillaryPageIds[NumSpids] = sub.getAncillaryPageId();
                                           if (n > 0)
                                              *s++ = '+';
                                           strn0cpy(s, I18nNormalizeLanguageCode(sub.languageCode), MAXLANGCODE1);
                                           s += strlen(s);
                                           if (n++ > 1)
                                              break;
                                           }
                                        }
                                    if (updatePrimaryDevice)
                                       cDevice::PrimaryDevice()->SetAvailableTrack(ttSubtitle, NumSpids, spids[NumSpids], slangs[NumSpids]);
*/
                                    NumSpids++;
                                    spids[NumSpids]= 0;
                                    }
                                 break;
                            case SI::ISO639LanguageDescriptorTag: {
                                 SI::ISO639LanguageDescriptor *ld = (SI::ISO639LanguageDescriptor *)d;
                                 dbgpatpmt(" '%s'", ld->languageCode);
//                                 strn0cpy(lang, I18nNormalizeLanguageCode(ld->languageCode), MAXLANGCODE1);
                                 }
                                 break;
                            default: ;
                            }
                          delete d;
                          }
                      if (dpid) {
                         if (NumDpids < MAXDPIDS) {
                            dpids[NumDpids] = dpid;
                            dtypes[NumDpids] = dtype;
//                            strn0cpy(dlangs[NumDpids], lang, sizeof(dlangs[NumDpids]));
//                            if (updatePrimaryDevice && Setup.UseDolbyDigital)
//                               cDevice::PrimaryDevice()->SetAvailableTrack(ttDolby, NumDpids, dpid, lang);
                            NumDpids++;
                            dpids[NumDpids]= 0;
                            }
                         }
                      }
                      break;
           default: ;
           }
         dbgpatpmt("\n");
//         if (updatePrimaryDevice) {
//            cDevice::PrimaryDevice()->EnsureAudioTrack(true);
//            cDevice::PrimaryDevice()->EnsureSubtitleTrack();
//            }
         }
     pmtVersion = Pmt.getVersionNumber();
     }
  else
     esyslog("ERROR: can't parse PMT");
  pmtSize = 0;
}

bool cPatPmtParser::GetVersions(int &PatVersion, int &PmtVersion) const
{
  PatVersion = patVersion;
  PmtVersion = pmtVersion;
  return patVersion >= 0 && pmtVersion >= 0;
}

// --- cNaluDumper ---------------------------------------------------------

cNaluDumper::cNaluDumper()
{
    LastContinuityOutput = -1;
    reset();
}

void cNaluDumper::reset()
{
    LastContinuityInput = -1;
    ContinuityOffset = 0;
    PesId = -1;
    PesOffset = 0;
    NaluFillState = NALU_NONE;
    NaluOffset = 0;
    History = 0xffffffff;
    DropAllPayload = false;
}

void cNaluDumper::ProcessPayload(unsigned char *Payload, int size, bool PayloadStart, sPayloadInfo &Info)
{
    Info.DropPayloadStartBytes = 0;
    Info.DropPayloadEndBytes = 0;
    int LastKeepByte = -1;

    if (PayloadStart)
    {
        History = 0xffffffff;
        PesId = -1;
        NaluFillState = NALU_NONE;
    }

    for (int i=0; i<size; i++) {
        History = (History << 8) | Payload[i];

        PesOffset++;
        NaluOffset++;

        bool DropByte = false;

        if (History >= 0x00000180 && History <= 0x000001FF)
        {
            // Start of PES packet
            PesId = History & 0xff;
            PesOffset = 0;
            NaluFillState = NALU_NONE;
        }
        else if (PesId >= 0xe0 && PesId <= 0xef // video stream
                 && History >= 0x00000100 && History <= 0x0000017F) // NALU start code
        {
            int NaluId = History & 0xff;
            NaluOffset = 0;
            NaluFillState = ((NaluId & 0x1f) == 0x0c) ? NALU_FILL : NALU_NONE;
        }

        if (PesId >= 0xe0 && PesId <= 0xef // video stream
            && PesOffset >= 1 && PesOffset <= 2)
        {
            Payload[i] = 0; // Zero out PES length field
        }

        if (NaluFillState == NALU_FILL && NaluOffset > 0) // Within NALU fill data
        {
            // We expect a series of 0xff bytes terminated by a single 0x80 byte.

            if (Payload[i] == 0xFF)
            {
                DropByte = true;
            }
            else if (Payload[i] == 0x80)
            {
                NaluFillState = NALU_TERM; // Last byte of NALU fill, next byte sets NaluFillEnd=true
                DropByte = true;
            }
            else // Invalid NALU fill
            {
                dsyslog("cNaluDumper: Unexpected NALU fill data: %02x", Payload[i]);
                NaluFillState = NALU_END;
                if (LastKeepByte == -1)
                {
                    // Nalu fill from beginning of packet until last byte
                    // packet start needs to be dropped
                    Info.DropPayloadStartBytes = i;
                }
            }
        }
        else if (NaluFillState == NALU_TERM) // Within NALU fill data
        {
            // We are after the terminating 0x80 byte
            NaluFillState = NALU_END;
            if (LastKeepByte == -1)
            {
                // Nalu fill from beginning of packet until last byte
                // packet start needs to be dropped
                Info.DropPayloadStartBytes = i;
            }
        }

        if (!DropByte)
            LastKeepByte = i; // Last useful byte
    }

    Info.DropAllPayloadBytes = (LastKeepByte == -1);
    Info.DropPayloadEndBytes = size-1-LastKeepByte;
}

bool cNaluDumper::ProcessTSPacket(unsigned char *Packet)
{
    bool HasAdaption = TsHasAdaptationField(Packet);
    bool HasPayload = TsHasPayload(Packet);

    // Check continuity:
    int ContinuityInput = TsContinuityCounter(Packet);
    if (LastContinuityInput >= 0)
    {
        int NewContinuityInput = HasPayload ? (LastContinuityInput + 1) & TS_CONT_CNT_MASK : LastContinuityInput;
        int Offset = (NewContinuityInput - ContinuityInput) & TS_CONT_CNT_MASK;
        if (Offset > 0)
            dsyslog("cNaluDumper: TS continuity offset %i", Offset);
        if (Offset > ContinuityOffset)
            ContinuityOffset = Offset; // max if packets get dropped, otherwise always the current one.
    }
    LastContinuityInput = ContinuityInput;

    if (HasPayload) {
        sPayloadInfo Info;
        int Offset = TsPayloadOffset(Packet);
        ProcessPayload(Packet + Offset, TS_SIZE - Offset, TsPayloadStart(Packet), Info);

        if (DropAllPayload && !Info.DropAllPayloadBytes)
        {
            // Return from drop packet mode to normal mode
            DropAllPayload = false;

            // Does the packet start with some remaining NALU fill data?
            if (Info.DropPayloadStartBytes > 0)
            {
                // Add these bytes as stuffing to the adaption field.

                // Sample payload layout:
                // FF FF FF FF FF 80 00 00 01 xx xx xx xx
                //                   ^DropPayloadStartBytes

                TsExtendAdaptionField(Packet, Offset - 4 + Info.DropPayloadStartBytes);
            }
        }

        bool DropThisPayload = DropAllPayload;

        if (!DropAllPayload && Info.DropPayloadEndBytes > 0) // Payload ends with 0xff NALU Fill
        {
            // Last packet of useful data
            // Do early termination of NALU fill data
            Packet[TS_SIZE-1] = 0x80;
            DropAllPayload = true;
            // Drop all packets AFTER this one

            // Since we already wrote the 0x80, we have to make sure that
            // as soon as we stop dropping packets, any beginning NALU fill of next
            // packet gets dumped. (see DropPayloadStartBytes above)
        }

        if (DropThisPayload && HasAdaption)
        {
            // Drop payload data, but keep adaption field data
            TsExtendAdaptionField(Packet, TS_SIZE-4);
            DropThisPayload = false;
        }

        if (DropThisPayload)
        {
            return true; // Drop packet
        }
    }

    // Fix Continuity Counter and reproduce incoming offsets:
    int NewContinuityOutput = TsHasPayload(Packet) ? (LastContinuityOutput + 1) & TS_CONT_CNT_MASK : LastContinuityOutput;
    NewContinuityOutput = (NewContinuityOutput + ContinuityOffset) & TS_CONT_CNT_MASK;
    TsSetContinuityCounter(Packet, NewContinuityOutput);
    LastContinuityOutput = NewContinuityOutput;
    ContinuityOffset = 0;

    return false; // Keep packet
}

// --- cNaluStreamProcessor ---------------------------------------------------------

cNaluStreamProcessor::cNaluStreamProcessor()
{
    pPatPmtParser = NULL;
    vpid = -1;
    data = NULL;
    length = 0;
    tempLength = 0;
    tempLengthAtEnd = false;
    TotalPackets = 0;
    DroppedPackets = 0;
}

void cNaluStreamProcessor::PutBuffer(uchar *Data, int Length)
{
    if (length > 0)
        esyslog("cNaluStreamProcessor::PutBuffer: New data before old data was processed!");

    data = Data;
    length = Length;
}

uchar* cNaluStreamProcessor::GetBuffer(int &OutLength)
{
    if (length <= 0)
    {
        // Need more data - quick exit
        OutLength = 0;
        return NULL;
    }
    if (tempLength > 0) // Data in temp buffer?
    {
        if (tempLengthAtEnd) // Data is at end, copy to beginning
        {
            // Overlapping src and dst!
            for (int i=0; i<tempLength; i++)
                tempBuffer[i] = tempBuffer[TS_SIZE-tempLength+i];
        }
        // Normalize TempBuffer fill
        if (tempLength < TS_SIZE && length > 0)
        {
            int Size = min(TS_SIZE-tempLength, length);
            memcpy(tempBuffer+tempLength, data, Size);
            data += Size;
            length -= Size;
            tempLength += Size;
        }
        if (tempLength < TS_SIZE)
        {
            // All incoming data buffered, but need more data
            tempLengthAtEnd = false;
            OutLength = 0;
            return NULL;
        }
        // Now: TempLength==TS_SIZE
        if (tempBuffer[0] != TS_SYNC_BYTE)
        {
            // Need to sync on TS within temp buffer
            int Skipped = 1;
            while (Skipped < TS_SIZE && (tempBuffer[Skipped] != TS_SYNC_BYTE || (Skipped < length && data[Skipped] != TS_SYNC_BYTE)))
                Skipped++;
            esyslog("ERROR: skipped %d bytes to sync on start of TS packet", Skipped);
            // Pass through skipped bytes
            tempLengthAtEnd = true;
            tempLength = TS_SIZE - Skipped; // may be 0, thats ok
            OutLength = Skipped;
            return tempBuffer;
        }
        // Now: TempBuffer is a TS packet
        int Pid = TsPid(tempBuffer);
        if (pPatPmtParser)
        {
            if (Pid == 0)
                pPatPmtParser->ParsePat(tempBuffer, TS_SIZE);
            else if (Pid == pPatPmtParser->PmtPid())
                pPatPmtParser->ParsePmt(tempBuffer, TS_SIZE);
        }

        TotalPackets++;
        bool Drop = false;
        if (Pid == vpid || (pPatPmtParser && Pid == pPatPmtParser->Vpid() && pPatPmtParser->Vtype() == 0x1B))
            Drop = NaluDumper.ProcessTSPacket(tempBuffer);
        if (!Drop)
        {
            // Keep this packet, then continue with new data
            tempLength = 0;
            OutLength = TS_SIZE;
            return tempBuffer;
        }
        // Drop TempBuffer
        DroppedPackets++;
        tempLength = 0;
    }
    // Now: TempLength==0, just process data/length

    // Pointer to processed data / length:
    uchar *Out = data;
    uchar *OutEnd = Out;

    while (length >= TS_SIZE)
    {
        if (data[0] != TS_SYNC_BYTE) {
            int Skipped = 1;
            while (Skipped < length && (data[Skipped] != TS_SYNC_BYTE || (length - Skipped > TS_SIZE && data[Skipped + TS_SIZE] != TS_SYNC_BYTE)))
                Skipped++;
            esyslog("ERROR: skipped %d bytes to sync on start of TS packet", Skipped);

            // Pass through skipped bytes
            if (OutEnd != data)
                memcpy(OutEnd, data, Skipped);
            OutEnd += Skipped;
            continue;
        }
        // Now: Data starts with complete TS packet

        int Pid = TsPid(data);
        if (pPatPmtParser)
        {
            if (Pid == 0)
                pPatPmtParser->ParsePat(data, TS_SIZE);
            else if (Pid == pPatPmtParser->PmtPid())
                pPatPmtParser->ParsePmt(data, TS_SIZE);
        }

        TotalPackets++;
        bool Drop = false;
        if (Pid == vpid || (pPatPmtParser && Pid == pPatPmtParser->Vpid() && pPatPmtParser->Vtype() == 0x1B))
            Drop = NaluDumper.ProcessTSPacket(data);
        if (!Drop)
        {
            if (OutEnd != data)
                memcpy(OutEnd, data, TS_SIZE);
            OutEnd += TS_SIZE;
        }
        else
        {
            DroppedPackets++;
        }
        data += TS_SIZE;
        length -= TS_SIZE;
    }
    // Now: Less than a packet remains.
    if (length > 0)
    {
        // copy remains into temp buffer
        memcpy(tempBuffer, data, length);
        tempLength = length;
        tempLengthAtEnd = false;
        length = 0;
    }
    OutLength = (OutEnd - Out);
    return OutLength > 0 ? Out : NULL;
}
